#ifndef PDQTYPES_H
#define PDQTYPES_H
#include <stdio.h>

namespace facebook {
namespace pdq {
namespace hashing {

using Hash8 = unsigned char;
using Hash16 = unsigned short;

} // namespace hashing
} // namespace pdq
} // namespace facebook

#endif // PDQTYPES_H
