#ifndef TORBEN_H
#define TORBEN_H
/*
 * The following code is public domain.
 * Algorithm by Torben Mogensen, implementation by N. Devillard.
 * This code in public domain.
 */

namespace facebook {
namespace pdq {
namespace hashing {
float torben(float m[], int n);
} // namespace hashing
} // namespace pdq
} // namespace facebook

#endif // TORBEN_H
