#ifndef TORBEN_H
#define TORBEN_H
/*
 * The following code is public domain.
 * Algorithm by Torben Mogensen, implementation by N. Devillard.
 * This code in public domain.
 */

float torben(float m[], int n);

#endif // TORBEN_H
